import logging
import re

from pathlib import Path
from .constants import HTTP_NO_PORT_REGEX


BASE_DIR = Path(__file__).resolve().parent
__version__ = "1.1.3"
__name__ = "spid_sp_test"
logger = logging.getLogger(__name__)


class AbstractSpidCheck(object):
    def __init__(self, *args, **kwargs):
        self.results = []
        self.errors = []
        self.warnings = []
        self.logger = logger
        self.error_counter = 0
        self.verify_ssl = kwargs.get("verify_ssl", False)
        self.category = ""

    def report_to_dict(self):
        res = {self.category: {self.__class__.__name__: self.results}}
        return res

    def is_ok(self, msg):
        if not self.error_counter:
            # self.handle_result(
            # "info",
            # msg,
            # method = method or msg
            # )
            return True
        else:
            self.error_counter = 0
            return False

    def handle_result(
        self,
        level: str,
        title: str,
        description: str = "",
        traceback: str = None,
        references: list = [],
        method: str = "",
        test_id: str = "",
    ):
        msg = f"{title}"
        getattr(self.logger, level, "debug")(f"{method}: {msg}")
        value = f"{description}" if not traceback else f"{description}: {traceback }"

        data = {
            "test_id": test_id,
            "test": title,
            "value": value.decode() if isinstance(value, bytes) else value,
            "references": references,
            "method": method,
        }

        if level not in ("error", "debug", "critical", "warning"):
            # here report as json
            data["result"] = "success"
            self.results.append(data)
        elif level in ("error", "critical"):
            self.handle_error(title, description, traceback)
        elif level == "warning":
            data["result"] = "warning"
            self.results.append(data)
            self.warnings.append(data)

    def handle_error(
        self,
        error_message,
        description="",
        traceback: str = None,
        references: list = [],
        method: str = "",
        test_id: str = "",
    ):
        self.logger.error(error_message)
        self.error_counter += 1
        # here report as json
        value = f"{description}" if not traceback else f"{description}: {traceback }"
        data = {
            "test_id": test_id,
            "result": "failure",
            "test": error_message,
            "value": value.decode() if isinstance(value, bytes) else value,
            "references": references,
            "method": method,
        }
        self.errors.append(data)
        self.results.append(data)

    def _assert(
        self,
        check: bool,
        error_message: str,
        description="",
        traceback: str = None,
        level: str = "info",
        **kwargs,
    ):
        if level == "warning":
            # overwrite to info is warning doens't happen.
            if check:
                level = "info"
        elif not check:
            level = "error"
        self.handle_result(level, error_message, description, traceback, **kwargs)

    def _assertTrue(self, *args, **kwargs):
        self._assert(*args, **kwargs)

    def _assertFalse(self, check, *args, **kwargs):
        self._assert(not check, *args, **kwargs)

    def _assertIsValidHttpsUrl(self, check, *args, **kwargs):
        self._assert(re.match("https://", check if check else ""), *args, **kwargs)

    def _assertHttpUrlWithoutPort(self, check, *args, **kwargs):
        self._assert(
            re.match(HTTP_NO_PORT_REGEX, check if check else ""), *args, **kwargs
        )

    def _assertIsValidHttpUrl(self, check, *args, **kwargs):
        self._assert(re.match("https?://", check if check else ""), *args, **kwargs)

    def handle_init_errors(self, method, description, traceback=""):
        self._assertTrue(
            False,
            method,
            description=description,
            traceback=traceback,
            method=method,
            test_id=[],
        )
        self.is_ok(method)
        raise Exception(traceback)

    # maybe useful .. one day ?!
    # idp_server = self.idp()
    # pysaml2 auth req object with signature check

    # SPID IdP not completely compliant to SAML2-core about Destination attribute
    # saml2.request:http://localhost:8088 not in ['http://localhost:54321/sso']
    # try:
    # req_obj = idp_server.parse_authn_request(self.authn_request_encoded,
    # BINDING_HTTP_POST)
    # except OtherError as e:
    # pass



import urllib


class SpidSpMetadataCheckAG(object):
    def test_extensions_public_ag(
        self,
        ext_types=[
            "//ContactPerson/Extensions/PublicServicesFullAggregator",
            "//ContactPerson/Extensions/PublicServicesLightAggregator",
            "//ContactPerson/Extensions/PrivateServicesFullAggregator",
            "//ContactPerson/Extensions/PrivateServicesLightAggregator",
            "//ContactPerson/Extensions/PublicServicesFullOperator",
            "//ContactPerson/Extensions/PublicServicesLightOperator",
        ],
        must=False,
    ):
        _method = f"{self.__class__.__name__}.test_extensions_public_ag"
        _data = dict(
            test_id="",
            references=[],
            method=_method,
        )
        for ext_type in ext_types:
            ctype = self.doc.xpath(ext_type)
            if must:
                self._assertTrue(
                    ctype, f"The {ext_type} element MUST be present", **_data
                )

            if ctype:
                self._assertFalse(
                    ctype[0].text,
                    f"The {ext_type.title()} element MUST be empty",
                    **_data,
                )

        return self.is_ok(_method)

    def test_entityid_qs(self):
        """The entityID MUST not contain the query-string part"""
        _method = f"{self.__class__.__name__}.test_entityid_qs"
        _data = dict(
            test_id="",
            references=[],
            method=_method,
        )
        entity_desc = self.doc.xpath("//EntityDescriptor")
        eid = entity_desc[0].get("entityID")

        qs = urllib.parse.splitquery(eid)

        self._assertFalse(
            qs[1],
            ("The entityID MUST not contain the query-string part"),
            description=eid,
            **_data,
        )
        return self.is_ok(_method)

    def test_entityid_contains(self, value=""):
        """The entityID MUST contain ..."""
        _method = f"{self.__class__.__name__}.test_entityid_contains"
        _data = dict(
            test_id="",
            references=[],
            method=_method,
        )
        entity_desc = self.doc.xpath("//EntityDescriptor")
        eid = entity_desc[0].get("entityID")
        self._assertTrue(
            value in eid,
            (f"The entityID MUST contain {value}"),
            description=eid,
            **_data,
        )
        return self.is_ok(_method)



from tempfile import NamedTemporaryFile
from ..authn_request import get_authn_request
from ..utils import saml_from_htmlform


class Base(object):
    def __init__(
        self,
        requests_session,
        authn_request_url,
        request_method: str = "GET",
        request_body: dict = {},
        request_content_type: str = "data",
    ):
        self.requests_session = requests_session
        self.authn_request_url = authn_request_url
        self.request_method = request_method or "GET"
        self.request_body = request_body
        self.request_content_type = request_content_type

    def request(self):
        """ """
        return self.authn_request_url

    def response(self, url, data: dict = {}, **kwargs):
        """ """
        req_dict = {
            "verify": kwargs.get("verify_ssl", 0),
            "allow_redirects": kwargs.get("allow_redirects", True),
        }
        if self.request_method.upper() == "POST":
            req_dict[self.request_content_type.lower()] = self.request_body
        method = getattr(self.requests_session, self.request_method.lower())

        res = method(url, **req_dict)
        return res


class SatosaSaml2Spid(Base):
    def request(self):
        """ """
        data = get_authn_request(
            self.authn_request_url, requests_session=self.requests_session
        )
        req = data["requests_session"]
        tmp_file = NamedTemporaryFile()
        if data["method"] == "post":
            form = {
                "SAMLRequest": data["SAMLRequest"],
                "RelayState": data.get("RelayState", ""),
            }
            res = req.post(data["action"], data=form, verify=0)
            tmp_file.write(res.content)
        else:
            res = req.get(data["SAMLRequest_redirect"], verify=0)
            tmp_file.write(res.headers["Location"])

        tmp_file.seek(0)
        _data = get_authn_request(f"file://{tmp_file.name}", requests_session=req)
        return _data

    def response(self, url, data: dict = {}, **kwargs):
        """ """
        res = self.requests_session.post(
            url, data=data, verify=0, allow_redirects=False
        )

        if res.status_code != 200:
            return res

        form = saml_from_htmlform(res.content.decode())
        url = form.pop("action")
        method = getattr(self.requests_session, form.pop("method"))
        next_res = method(url, data=dict(form), verify=0)
        return next_res



import base64
import datetime
import importlib
import json
import logging
import random
import requests
import os
import sys
import string


from copy import deepcopy
from jinja2 import Environment, Markup, FileSystemLoader, Template, select_autoescape
from lxml import etree

from saml2.sigver import CryptoBackendXmlSec1, XmlsecError
from spid_sp_test import BASE_DIR, AbstractSpidCheck
from spid_sp_test.authn_request import get_authn_request
from spid_sp_test.idp.settings import SAML2_IDP_CONFIG
from spid_sp_test.responses import settings
from spid_sp_test.utils import del_ns
from tempfile import NamedTemporaryFile

from .constants import NOSESINDEX_ACRS
from .utils import get_xmlsec1_bin, html_absolute_paths
from .utils import load_plugin

logger = logging.getLogger(__name__)


def stupid_rnd_string(N=32):
    return "".join(random.choice(string.ascii_lowercase) for _ in range(N))


def saml_rnd_id():
    return (
        f"_{stupid_rnd_string(8)}"
        f"-{stupid_rnd_string(4)}"
        f"-{stupid_rnd_string(4)}"
        f"-{stupid_rnd_string(4)}"
        f"-{stupid_rnd_string(12)}"
    )


class SpidSpResponse(object):
    def __init__(
        self,
        conf=None,
        response_attrs={},
        authnreq_attrs={},
        attributes={},
        template_path="./templates",
        status_codes=None,
    ):

        try:
            self.conf = deepcopy(
                settings.RESPONSE_TESTS[conf] or settings.RESPONSE_TESTS["1"]
            )
        except KeyError:
            raise Exception(f"Test {conf} doesn't exists")

        # rewrite, eg: response_mods
        if status_codes:
            self.conf["status_codes"] = self.status_codes

        self.attributes = attributes
        self.authnreq_attrs = authnreq_attrs
        self.response_attrs = response_attrs

        self.loader = Environment(
            loader=FileSystemLoader(searchpath=template_path),
            autoescape=select_autoescape(["xml"]),
        )
        self.template_name = self.conf.get("path", "base.xml")
        self.private_key = self.conf.get("sign_credentials", {}).get("privateKey")

    def render_attributes(self, attributes={}):
        """
        fill values to be released as identity attributes
        """
        attr_rendr_list = []
        attrs = attributes or self.attributes or settings.ATTRIBUTES

        for k, v in attrs.items():
            template = Template(settings.ATTRIBUTE_TMPL)
            attr_type = settings.ATTRIBUTES_TYPES.get(k, "string")
            attr_rendr = template.render(name=k, value=v, type=attr_type)
            attr_rendr_list.append(attr_rendr)
        return Markup("\n".join(attr_rendr_list))

    def render(self, user_attrs: dict = {}, data: dict = {}):
        template = self.loader.get_template(self.template_name)
        data = data or self.response_attrs
        data["Attributes"] = self.render_attributes(attributes=user_attrs)
        result = template.render(**data)
        logger.debug(f"Rendering response template {template}: {result}")
        return result

    def __str__(self):
        return self.conf


class SpidSpResponseCheck(AbstractSpidCheck):
    template_path = f"{BASE_DIR}/responses/templates/"

    def __init__(self, *args, **kwargs):
        super(SpidSpResponseCheck, self).__init__(*args, **kwargs)
        self.category = "response"

        self.template_path = kwargs.get("template_path", self.template_path)
        self.metadata_etree = kwargs.get("metadata_etree")

        self.authn_request_url = kwargs.get("authn_request_url")
        self.authn_request_data = {}

        # signing
        self.crypto_backend = CryptoBackendXmlSec1(
            xmlsec_binary=kwargs.get("xmlsec_binary") or get_xmlsec1_bin()
        )
        self.private_key_fpath = SAML2_IDP_CONFIG["key_file"]

        # tests
        self.tests = {}
        for i in kwargs.get("test_jsons", []):
            with open(i[0], "r") as json_data:
                self.tests.update(json.loads(json_data.read()))
        if not self.tests:
            self.tests.update(settings.RESPONSE_TESTS)
        self.test_names = kwargs.get("test_names") or self.tests.keys()

        self.attr_json = kwargs.get("attr_json")

        self.html_path = kwargs.get("html_path")
        self.no_send_response = kwargs.get("no_send_response")
        self.kwargs = kwargs
        self.status_codes = None

        self.authn_plugin = kwargs.get("authn_plugin")
        self.requests_session = kwargs.get("requests_session")

        self.request_method = kwargs.get("request_method")
        self.request_body = kwargs.get("request_body")
        self.request_content_type = kwargs.get("request_content_type")

    def get_acr(self):
        _acr = self.authnreq_etree.xpath("//RequestedAuthnContext/AuthnContextClassRef")
        if _acr:
            return _acr[0].text

    def get_acr_comparison(self):
        _acr = self.authnreq_etree.xpath("//RequestedAuthnContext")
        if _acr:
            return _acr[0].attrib.get("Comparison", "minimum")

    def load_user_attributes(self):
        self.requested_attrs_raw = self.metadata_etree.xpath(
            f"//SPSSODescriptor/AttributeConsumingService[@index={self.acs_index}]"
            "/RequestedAttribute"
        )
        self.requested_attrs = [
            i.attrib["Name"] for i in self.requested_attrs_raw
        ]

        # attributes
        if self.attr_json:
            with open(self.attr_json, "r") as json_data:
                self.user_attrs = json.loads(json_data.read())
        else:
            # returns ONLY the requested attributes shown in the metadata
            # otherwise it returns all the attributes (for test purpose)
            self.user_attrs = {
                i:settings.ATTRIBUTES[i] for i in self.requested_attrs
            } or settings.ATTRIBUTES

    def do_authnrequest(self):
        self.authn_request_data = get_authn_request(
            self.authn_request_url,
            authn_plugin=self.authn_plugin,
            requests_session=self.requests_session,
            request_method=self.request_method,
            request_body=self.request_body,
            request_content_type=self.request_content_type,
        )
        self.authnreq_etree = etree.fromstring(
            self.authn_request_data["SAMLRequest_xml"]
        )
        del_ns(self.authnreq_etree)

        self.issuer = self.kwargs.get("issuer", SAML2_IDP_CONFIG["entityid"])
        self.authnreq_attrs = self.authnreq_etree.xpath("/AuthnRequest")[0].attrib
        self.authnreq_issuer = self.authnreq_etree.xpath("/AuthnRequest/Issuer")[
            0
        ].attrib["NameQualifier"]

        now = datetime.datetime.utcnow()

        self.acs_index = self.authnreq_attrs.get('AttributeConsumingServiceIndex')
        self.acs_url = self.metadata_etree.xpath(
            f"//SPSSODescriptor/AssertionConsumerService[@index={self.acs_index}]"
        )[0].attrib["Location"]
        self.acr = self.get_acr()

        if self.acr in NOSESINDEX_ACRS:
            _session_index = None
        else:
            _session_index = saml_rnd_id()
        self.response_attrs = {
            "ResponseID": saml_rnd_id(),
            "AuthnRequestID": self.authnreq_attrs["ID"],
            "IssueInstant": self.authnreq_attrs["IssueInstant"],
            "NotOnOrAfter": (now + datetime.timedelta(minutes=5)).strftime(
                "%Y-%m-%dT%H:%M:%SZ"
            ),
            "AssertionConsumerURL": self.authnreq_attrs.get(
                "AssertionConsumerURL", self.acs_url
            ),
            "NameIDNameQualifier": settings.DEFAULT_RESPONSE["NameIDNameQualifier"],
            "NameID": "that-transient-opaque-value",
            "AssertionID": saml_rnd_id(),
            "AuthnIstant": datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"),
            "SessionIndex": _session_index,
            "Issuer": self.issuer,
            "Audience": self.authnreq_issuer,
            "AuthnContextClassRef": (
                self.acr or settings.DEFAULT_RESPONSE["AuthnContextClassRef"]
            ),
            "IssueInstantMillis": now.strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
            "sign_response": settings.DEFAULT_RESPONSE["sign_response"],
            "sign_assertion": settings.DEFAULT_RESPONSE["sign_assertion"],
        }
        self.relay_state = self.kwargs.get("relay_state")

    def sign(self, xmlstr, assertion=True, response=True, key_file=None):
        """
        Sign an XML statement.
        """
        signature_node = Template(settings.SIGNATURE_TMPL)
        key_file = key_file or self.private_key_fpath
        com_list = [self.crypto_backend.xmlsec, "--sign", "--privkey-pem", key_file]

        asser_placeholder = "<!-- Assertion Signature here -->"
        if assertion and asser_placeholder in xmlstr:
            assertion_id = self.response_attrs.get("AssertionID")
            if assertion_id:
                ref_data = {"ReferenceURI": f"#{assertion_id}"}
            else:
                ref_data = {}
            value = signature_node.render(**ref_data)
            xmlstr = xmlstr.replace(asser_placeholder, value)

            with NamedTemporaryFile(suffix=".xml") as ntf:
                ntf.write(xmlstr.encode())
                ntf.seek(0)
                _com_list = [
                    "--id-attr:ID",
                    "urn:oasis:names:tc:SAML:2.0:assertion:Assertion",
                    # '--node-id',
                    # f'{self.response_attrs["AssertionID"]}'
                ]
                _com = com_list + _com_list
                p_out, p_err, xmlstr = self.crypto_backend._run_xmlsec(_com, [ntf.name])
                xmlstr = xmlstr.decode()

        sign_placeholder = "<!-- Response Signature here -->"
        if response and sign_placeholder in xmlstr:
            response_id = self.response_attrs.get("ResponseID")
            if response_id:
                ref_data = {"ReferenceURI": f"#{response_id}"}
            else:
                ref_data = {}
            value = signature_node.render(**ref_data)
            xmlstr = xmlstr.replace(sign_placeholder, value)
            with NamedTemporaryFile(suffix=".xml") as ntf:
                ntf.write(xmlstr.encode())
                ntf.seek(0)
                _com_list = [
                    "--id-attr:ID",
                    "urn:oasis:names:tc:SAML:2.0:protocol:Response",
                    # '--node-id',
                    # f'{self.response_attrs["ResponseID"]}'
                ]
                _com = com_list + _com_list
                p_out, p_err, xmlstr = self.crypto_backend._run_xmlsec(_com, [ntf.name])
                xmlstr = xmlstr.decode()

        return xmlstr.decode() if isinstance(xmlstr, bytes) else xmlstr

    def load_test(self, test_name=None, attributes={}, response_attrs={}):
        spid_response = SpidSpResponse(
            test_name,
            authnreq_attrs=self.authnreq_attrs,
            attributes=attributes,
            response_attrs=response_attrs or self.response_attrs,
            template_path=self.template_path,
            status_codes=self.status_codes,
        )
        conf = settings.RESPONSE_TESTS[test_name]
        if conf.get("response"):
            for k, v in conf["response"].items():
                logger.debug(f"Test {test_name}: overwriting {k} with {v}")
                spid_response.response_attrs[k] = v

        # response dinamyc mods and rewrites (plugins)
        if conf.get("response_mods"):
            for mod_func in conf["response_mods"]:
                n1, _, n2 = mod_func.rpartition(".")
                module = importlib.import_module(n1)
                func = getattr(module, n2)
                func(self, spid_response)

        return spid_response

    def check_response(self, res, msg: str, attendeds=[], test_id=""):
        if res.status_code in attendeds:
            status = True
        else:
            status = False
        status_code = f"[http status_code: {res.status_code}]"
        self._assertTrue(
            status,
            msg,
            method=f"{self.__class__.__name__}.check_response",
            description=status_code,
            references=[],
            test_id=test_id,
        )
        return status, status_code

    def dump_html_response(self, fname, description, result, content):
        try:
            content = html_absolute_paths(content, self.acs_url)
        except Exception as e:
            logger.critical(
                f"Something went wrong making absolute links in html content: {e}"
            )

        content = content.decode() if isinstance(content, bytes) else content
        head = f"<!-- {description} -->\n\n" f"<!-- {result} -->\n\n"
        os.makedirs(self.html_path, exist_ok=True)
        with open(f"{self.html_path}/{fname}.html", "w") as f:
            f.write(head)
            f.write(content)

    def send_response(self, xmlstr):
        data = {
            "RelayState": self.authn_request_data.get("RelayState", "/"),
            "SAMLResponse": base64.b64encode(xmlstr.encode()),
        }
        url = self.authnreq_attrs.get("AssertionConsumerURL", self.acs_url)
        ua = self.requests_session  # self.authn_request_data["requests_session"]
        if self.authn_plugin:
            func = load_plugin(self.authn_plugin)
            res = func(ua, self.authn_request_url).response(url, data)
        else:
            res = ua.post(url, data=data, allow_redirects=True)
        msg = f"Response http status code [{res.status_code}]: {res.content.decode()}"
        self.logger.debug(msg)
        return res

    def test_profile_spid_sp(self):
        for i in self.test_names:
            self.do_authnrequest()
            self.load_user_attributes()
            response_obj = self.load_test(test_name=i)
            test_display_desc = response_obj.conf["description"]
            msg = f'Response [{i}] "{test_display_desc}"'
            xmlstr = response_obj.render(user_attrs=self.user_attrs)
            try:
                result = self.sign(xmlstr, key_file=response_obj.private_key)
            except XmlsecError as e:
                logger.error(f"{msg}: Exception during xmlsec signature ({e})")
                logger.debug("{xmlstr}")
                break

            logger.debug(result)

            if not self.no_send_response and not self.authn_request_data.get(
                "requests_session"
            ):
                logger.critical(
                    "Responsens will be not sent. "
                    "You MUST adopt a http authn url to send "
                    "responses interactively."
                )
                sys.exit(1)

            if self.no_send_response:
                print(f"<!-- {test_display_desc} -->\n{result}")

            else:
                res = self.send_response(result)
                status, status_msg = self.check_response(
                    res,
                    msg=f"Test [{i}] {test_display_desc}",
                    attendeds=response_obj.conf["status_codes"],
                    test_id=i,
                )
                if self.html_path:
                    self.dump_html_response(
                        f"{i}_{status}",
                        response_obj.conf["description"],
                        result,
                        res.content.decode(),
                    )

            # reset session
            self.requests_session = requests.Session()

        self.is_ok(f"{self.__class__.__name__}")



import logging

from pathlib import PosixPath
from spid_compliant_certificates.validator.validate import validate

logger = logging.getLogger(__name__)


def _get_tests(report:dict):
    _res = []
    for i in report['tests']:
        for e in i['checks']:
            _res.append(e)
    return _res


def check_certificate(cert_path:str, sector:str="public"):
    _cert = PosixPath(cert_path)
    _val = validate(_cert, sector)

    report = _val.as_dict()
    return _get_tests(report)


if __name__ == '__main__':
    cert_path = PosixPath("src/spid_sp_test/idp/public.cert")
    res = check_certificate(cert_path)



import logging

from spid_sp_test.response import SpidSpResponse, SpidSpResponseCheck
from .settings import HTTP_STATUS_ERROR_CODES

logger = logging.getLogger(__name__)


def dynamic_acr(check: SpidSpResponseCheck, response_obj: SpidSpResponse, **kwargs):
    if check.get_acr() != check.response_attrs["AuthnContextClassRef"]:
        response_obj.conf["status_codes"] = HTTP_STATUS_ERROR_CODES
        try:
            level_sp = int(check.get_acr()[-1])
            level_idp = int(check.response_attrs["AuthnContextClassRef"][-1])
            if level_idp > level_sp and check.get_acr_comparison() == "minimum":
                logger.debug(f"Spid level {level_idp} > {level_sp} and Comparison is {check.get_acr_comparison()} => Expecting OK ")
                response_obj.conf["status_codes"] = [200]
        except Exception:
            pass



# Copyright 2019 AgID - Agenzia per l'Italia Digitale
#
# Licensed under the EUPL, Version 1.2 or - as soon they will be approved by
# the European Commission - subsequent versions of the EUPL (the "Licence").
#
# You may not use this work except in compliance with the Licence.
#
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS" basis, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# Licence for the specific language governing permissions and limitations
# under the Licence.

import re

# iso8601 fully compliant regex
UTC_STRING = r"^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-(3[01]|0[1-9]|[12][0-9])T(2[0-3]|[01][0-9]):([0-5][0-9]):([0-5][0-9])(\.[0-9]+)?(Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])?$"  # noqa

UTC_REGEXP = re.compile(UTC_STRING)

HTTP_NO_PORT_REGEX = re.compile(
    r"^(?:http)s?://"  # http:// or https://
    r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)"  # domain...
    # r'|localhost|' #localhost...
    # r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}' # ...or ip
    r")"
    # r'(?::\d+)?' # optional port
    r"(?:/?|[/?]\S+)$",
    re.IGNORECASE,
)

_SPID_LEVEL_23 = r"(https:\/\/www\.spid\.gov\.it\/)SpidL[2-3]"  # noqa
SPID_LEVEL_23 = re.compile(_SPID_LEVEL_23)

_SPID_LEVEL_ALL = r"(https:\/\/www\.spid\.gov\.it\/)SpidL[1-3]"  # noqa
SPID_LEVEL_ALL = re.compile(_SPID_LEVEL_ALL)

BOOLEAN_TRUE = [
    "true",
    "1",
]

ALLOWED_BINDINGS = [
    "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST",
    "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect",
]

ALLOWED_SINGLELOGOUT_BINDINGS = [
    "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST",
    "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect",
    "urn:oasis:names:tc:SAML:2.0:bindings:SOAP",
]

ALLOWED_FORMATS = [
    "urn:oasis:names:tc:SAML:2.0:attrname-format:basic",
    "urn:oasis:names:tc:SAML:2.0:attrname-format:uri",
]

ALLOWED_STATUS_CODES = [
    "urn:oasis:names:tc:SAML:2.0:status:AuthnFailed",
    "urn:oasis:names:tc:SAML:2.0:status:NoAuthnContext",
    "urn:oasis:names:tc:SAML:2.0:status:NoPassive",
    "urn:oasis:names:tc:SAML:2.0:status:RequestDenied",
    "urn:oasis:names:tc:SAML:2.0:status:RequestUnsupported",
    "urn:oasis:names:tc:SAML:2.0:status:Requester",
    "urn:oasis:names:tc:SAML:2.0:status:Success",
    "urn:oasis:names:tc:SAML:2.0:status:VersionMismatch",
]

ALLOWED_XMLDSIG_ALGS = [
    "http://www.w3.org/2001/04/xmldsig-more#ecdsa-sha256",
    "http://www.w3.org/2001/04/xmldsig-more#ecdsa-sha384",
    "http://www.w3.org/2001/04/xmldsig-more#ecdsa-sha512",
    "http://www.w3.org/2001/04/xmldsig-more#hmac-sha256",
    "http://www.w3.org/2001/04/xmldsig-more#hmac-sha384",
    "http://www.w3.org/2001/04/xmldsig-more#hmac-sha512",
    "http://www.w3.org/2001/04/xmldsig-more#rsa-sha256",
    "http://www.w3.org/2001/04/xmldsig-more#rsa-sha384",
    "http://www.w3.org/2001/04/xmldsig-more#rsa-sha512",
]

ALLOWED_DGST_ALGS = [
    "http://www.w3.org/2001/04/xmlenc#sha256",
    "http://www.w3.org/2001/04/xmlenc#sha384",
    "http://www.w3.org/2001/04/xmlenc#sha512",
]

SPID_ATTRIBUTES = [
    "address",
    "companyName",
    "companyFiscalNumber",
    "countyOfBirth",
    "dateOfBirth",
    "digitalAddress",
    "email",
    "expirationDate",
    "familyName",
    "fiscalNumber",
    "gender",
    "idCard",
    "ivaCode",
    "mobilePhone",
    "name",
    "placeOfBirth",
    "registeredOffice",
    "spidCode",
    "domicileStreetAddress",
    "domicilePostalCode",
    "domicileMunicipality",
    "domicileProvince",
    "domicileNation",
]

CIE_ATTRIBUTES = ["dateOfBirth", "familyName", "fiscalNumber", "name"]

FICEP_MIN_ATTRIBUTES = [
    "fiscalNumber",
    "spidCode",
    "name",
    "familyName",
    "dateOfBirth",
]

FICEP_FULL_ATTRIBUTES = [
    *FICEP_MIN_ATTRIBUTES,
    "placeOfBirth",
    "address",
    "gender",
]

FICEP_MINIMUM_SET_SERVICENAME = "eIDAS Natural Person Minimum Attribute Set"
FICEP_FULL_SET_SERVICENAME = "eIDAS Natural Person Full Attribute Set"

SPID_LEVELS = [
    "https://www.spid.gov.it/SpidL1",
    "https://www.spid.gov.it/SpidL2",
    "https://www.spid.gov.it/SpidL3",
    "urn:oasis:names:tc:SAML:2.0:ac:classes:SpidL1",
    "urn:oasis:names:tc:SAML:2.0:ac:classes:SpidL2",
    "urn:oasis:names:tc:SAML:2.0:ac:classes:SpidL3",
]

NOSESINDEX_ACRS = [
    "https://www.spid.gov.it/SpidL2",
    "https://www.spid.gov.it/SpidL3",
    "urn:oasis:names:tc:SAML:2.0:ac:classes:SpidL2",
    "urn:oasis:names:tc:SAML:2.0:ac:classes:SpidL3",
]

ONE_MONTH = 30
SIX_MONTHS = 182
ONE_YEAR = 365

MINIMUM_CERTIFICATE_LENGHT = 2048  # type: int
DESIRED_CERTIFICATE_LENGHT = 3072  # type: int

NODE_NAME = "urn:oasis:names:tc:SAML:2.0:assertion:Assertion"
EMAIL_REGEXP = r"^(\w|\.|\_|\-)+[@](\w|\_|\-|\.)+[.]\w{2,6}$"


ISO3166_CODES = [
    "AD",
    "AE",
    "AF",
    "AG",
    "AI",
    "AL",
    "AM",
    "AO",
    "AQ",
    "AR",
    "AS",
    "AT",
    "AU",
    "AW",
    "AX",
    "AZ",
    "BA",
    "BB",
    "BD",
    "BE",
    "BF",
    "BG",
    "BH",
    "BI",
    "BJ",
    "BL",
    "BM",
    "BN",
    "BO",
    "BQ",
    "BR",
    "BS",
    "BT",
    "BV",
    "BW",
    "BY",
    "BZ",
    "CA",
    "CC",
    "CD",
    "CF",
    "CG",
    "CH",
    "CI",
    "CK",
    "CL",
    "CM",
    "CN",
    "CO",
    "CR",
    "CU",
    "CV",
    "CW",
    "CX",
    "CY",
    "CZ",
    "DE",
    "DJ",
    "DK",
    "DM",
    "DO",
    "DZ",
    "EC",
    "EE",
    "EG",
    "EH",
    "ER",
    "ES",
    "ET",
    "FI",
    "FJ",
    "FK",
    "FM",
    "FO",
    "FR",
    "GA",
    "GB",
    "GD",
    "GE",
    "GF",
    "GG",
    "GH",
    "GI",
    "GL",
    "GM",
    "GN",
    "GP",
    "GQ",
    "GR",
    "GS",
    "GT",
    "GU",
    "GW",
    "GY",
    "HK",
    "HM",
    "HN",
    "HR",
    "HT",
    "HU",
    "ID",
    "IE",
    "IL",
    "IM",
    "IN",
    "IO",
    "IQ",
    "IR",
    "IS",
    "IT",
    "JE",
    "JM",
    "JO",
    "JP",
    "KE",
    "KG",
    "KH",
    "KI",
    "KM",
    "KN",
    "KP",
    "KR",
    "KW",
    "KY",
    "KZ",
    "LA",
    "LB",
    "LC",
    "LI",
    "LK",
    "LR",
    "LS",
    "LT",
    "LU",
    "LV",
    "LY",
    "MA",
    "MC",
    "MD",
    "ME",
    "MF",
    "MG",
    "MH",
    "MK",
    "ML",
    "MM",
    "MN",
    "MO",
    "MP",
    "MQ",
    "MR",
    "MS",
    "MT",
    "MU",
    "MV",
    "MW",
    "MX",
    "MY",
    "MZ",
    "NA",
    "NC",
    "NE",
    "NF",
    "NG",
    "NI",
    "NL",
    "NO",
    "NP",
    "NR",
    "NU",
    "NZ",
    "OM",
    "PA",
    "PE",
    "PF",
    "PG",
    "PH",
    "PK",
    "PL",
    "PM",
    "PN",
    "PR",
    "PS",
    "PT",
    "PW",
    "PY",
    "QA",
    "RE",
    "RO",
    "RS",
    "RU",
    "RW",
    "SA",
    "SB",
    "SC",
    "SD",
    "SE",
    "SG",
    "SH",
    "SI",
    "SJ",
    "SK",
    "SL",
    "SM",
    "SN",
    "SO",
    "SR",
    "SS",
    "ST",
    "SV",
    "SX",
    "SY",
    "SZ",
    "TC",
    "TD",
    "TF",
    "TG",
    "TH",
    "TJ",
    "TK",
    "TL",
    "TM",
    "TN",
    "TO",
    "TR",
    "TT",
    "TV",
    "TW",
    "TZ",
    "UA",
    "UG",
    "UM",
    "US",
    "UY",
    "UZ",
    "VA",
    "VC",
    "VE",
    "VG",
    "VI",
    "VN",
    "VU",
    "WF",
    "WS",
    "YE",
    "YT",
    "ZA",
    "ZM",
    "ZW",
]

XML_NAMESPACES = {"spid": "https://spid.gov.it/saml-extensions"}
